# coding: utf-8
from itertools import imap
import abc
import operator

from django.db import models
from django.db.models.fields.related import RelatedField
from django.db.models.fields.reverse_related import ForeignObjectRel
from django.db.models.manager import Manager
from django.db.models import Q
from django.utils.encoding import force_text
from m3.actions.exceptions import ApplicationLogicException
from xlsxwriter import Workbook

from educommon.report.constructor.utils import get_field
from educommon.report.constructor.utils import get_field_value_by_display
from educommon.utils.misc import cached_property

from ... import constants
from ...base import ColumnDescriptor
from ...exceptions import DataSourceParamsNotFound
from ...exceptions import FilterError
from ...exceptions import ReportConstructorException
from ...models import ReportFilter
from ...models import ReportFilterGroup
from ...models import ReportSorting
from ...models import ReportTemplate
from ...registries import registry
from ...utils import get_columns_hierarchy


def _get_filter_values(report_filter, column_descriptor):
    u"""Возвращает значение из фильтра, преобразованное к объекту Python.

    Тип возвращаемого значения определяется оператором сравнения, который
    указан в фильтре:

        * **Пусто**: возвращается значения ``None``.
        * **Равно одному и**: возвращает кортеж из элементов с типом,
          соответствующим типу поля, по которому выполняется фильтрация данных.
        * **Между**: возвращает кортеж из двух элементов с типом,
          соответствующим типу поля, по которому выполняется фильтрация данных.
        * Во всех остальных случаях возвращает значение фильтра,
          преобразованное к типу данных того столбца, по которому выполняется
          фильтрация.

    .. note::

       Под значением фильтра подразумевается значение поля ``values`` модели
       :class:`~educommon.report.constructor.models.ReportFilter`.

    :param report_filter: Фильтр шаблона отчетов.
    :type report_filter: educommon.report.constructor.models.ReportFilter

    :param column_descriptor: Дескриптор поля модели, входящего в состав
        столбца отчета.
    ::type column_descriptor:
        educommon.report.constructor.base.ColumnDescriptor
    """

    def to_python(value):
        u"""Возвращает значение из фильтра, приведенное к объекту Python.

        В параметрах фильтра значение хранится в виде текстовой строки.
        """
        if column_descriptor.data_type == constants.CT_CHOICES:
            return get_field_value_by_display(column_descriptor.field, value)
        else:
            return column_descriptor.field.formfield().to_python(value)
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    if report_filter.operator == constants.IS_NULL:
        result = None

    elif report_filter.operator == constants.IN:
        result = tuple(
            to_python(value)
            for value in report_filter.values
        )

    elif report_filter.operator == constants.BETWEEN:
        result = tuple(
            to_python(value)
            for value in report_filter.values
        )

        if len(result) != 2:
            raise FilterError(
                report_filter, u'не правильно заданы границы диапазона'
            )

    elif report_filter.values:
        result = to_python(report_filter.values[0])

    else:
        raise FilterError(
            report_filter, u'не указано значение для сравнения.'
        )

    return result


class _FilterBuilder(object):

    u"""Построитель фильтров для выборки данных из БД."""

    def __init__(self, report_filter, column_descriptor):
        u"""Инициализация экземпляра.

        :param report_filter: Фильтр шаблона отчетов.
        :type report_filter: educommon.report.constructor.models.ReportFilter

        :param column_descriptor: Дескриптор поля модели, входящего в состав
            столбца отчета.
        ::type column_descriptor:
            educommon.report.constructor.base.ColumnDescriptor
        """
        assert isinstance(report_filter, ReportFilter), type(report_filter)
        assert isinstance(column_descriptor, ColumnDescriptor), type(
            column_descriptor
        )

        self._report_filter = report_filter
        self._column_descriptor = column_descriptor

    def _get_lookup(self, field_lookup):
        u"""Возвращает lookup-выражение, соответствующее параметрам фильтра.

        При формировании lookup-выражения учитываются параметры фильтра
        ``case_sensitive``, ``operator``, а также тип данных столбца, по
        которому выполняется фильтрация.

        :param str field_lookup: lookup-выражение для доступа к значению поля
        """
        if (
            not self._report_filter.case_sensitive and
            self._column_descriptor.data_type == constants.CT_TEXT
        ):
            field_lookup += {
                constants.EQ: '__iexact',
                constants.CONTAINS: '__icontains',
                constants.STARTS_WITH: '__istartswith',
                constants.ENDS_WITH: '__iendswith',
                constants.IN: '__lower__in',
            }.get(self._report_filter.operator, '')

        else:
            field_lookup += {
                constants.LE: '__lte',
                constants.LT: '__lt',
                constants.GT: '__gt',
                constants.GE: '__gte',
                constants.CONTAINS: '__contains',
                constants.STARTS_WITH: '__startswith',
                constants.ENDS_WITH: '__endswith',
                constants.BETWEEN: '__range',
                constants.IN: '__in',
            }.get(self._report_filter.operator, '')

        return field_lookup

    def _get_values(self):
        if self._report_filter.operator == constants.IS_NULL:
            values = True

        elif (
            not self._report_filter.case_sensitive and
            self._column_descriptor.data_type == constants.CT_TEXT and
            self._report_filter.operator == constants.IN
        ):
            values = tuple(v.lower() for v in values)

        else:
            values = _get_filter_values(
                self._report_filter, self._column_descriptor
            )

        return values

    def get_orm_filter(self):
        u"""Возвращает условия выборки из шаблона отчета.

        :rtype: django.db.models.Q
        """
        field_lookup = self._column_descriptor.full_lookup

        filter_values = self._get_values()

        if self._report_filter.operator == constants.IS_NULL:
            if self._column_descriptor.data_type == constants.CT_TEXT:
                result = Q(
                    Q(**{field_lookup + '__isnull': True}) |
                    Q(**{field_lookup: u''})
                )
            else:
                result = Q(**{field_lookup + '__isnull': True})

        else:
            try:
                result = Q(**{
                    self._get_lookup(field_lookup): filter_values,
                })
            except ValueError as error:
                raise FilterError(self._report_filter, unicode(error))

        if self._report_filter.exclude:
            result = ~result

        return result


class _FilterGroupBuilder(object):

    u"""Построитель Q-выражений и фильтрующих функций для групп фильтров."""

    def __init__(self, data_filterer, filter_group):
        u"""Инициализация экземпляра класса.

        :param data_filterer: Фильтратор данных отчета.
        :type data_filterer: _DataFilterer

        :param filter_group: Группа фильтров.
        :type filter_group:
            educommon.report.constructor.models.ReportFilterGroup
        """
        assert isinstance(data_filterer, _DataFilterer), type(data_filterer)
        assert isinstance(filter_group, ReportFilterGroup), type(
            filter_group
        )

        self._data_filterer = data_filterer
        self._filter_group = filter_group

    def _get_nested_orm_filters(self):
        for nested_group in self._filter_group.nested_groups.all():
            if any(
                report_filter
                for report_filter in self._data_filterer.itervalues()
                if report_filter.group_id == nested_group.id
            ):
                yield nested_group.get_orm_filter()

        data_source = self._data_filterer.data_source_descriptor

        for report_filter in self._data_filterer.filters_by_id.itervalues():
            if report_filter.group_id == self._filter_group.id:
                column_descriptor = data_source.get_column_descriptor(
                    report_filter.column.name
                )
                filter_builder = _FilterBuilder(
                    report_filter, column_descriptor
                )
                yield filter_builder.get_orm_filter()

    def get_orm_filter(self):
        u"""Возвращает Q-выражение для данной группы фильтров.

        :rtype: django.db.models.Q
        """
        operators = {
            ReportFilterGroup.OPERATOR_AND: operator.and_,
            ReportFilterGroup.OPERATOR_OR: operator.or_,
        }

        if self._filter_group.operator not in operators:
            raise ApplicationLogicException(
                u'Неподдерживаемый оператор: {}'
                .format(self._filter_group.operator)
            )

        return reduce(
            operators[self._filter_group.operator],
            self._get_nested_orm_filters()
        )


class _DataFilterer(object):

    u"""Фильтратор данных отчета.

    Фильтрация данных выполняется в два этапа:

        1. Наложение фильтров на запрос.
        2. Фильтрация полученных данных в приложении.
    """

    def __init__(self, report_template, data_source_descriptor,
                 report_columns):
        u"""Инициализация экземпляра класса.

        :param report_template: Шаблон фильтра.
        :type filter_group: educommon.report.constructor.models.ReportTemplate

        :param data_source_descriptor: Дескриптор источника данных..
        :type data_source_descriptor:
            educommon.report.constructor.base.ModelDataSourceDescriptor

        :param report_columns: Колонки отчета, упорядоченные по порядковому
            номеру.
        :type: tuple
        """
        assert isinstance(report_template, ReportTemplate), type(
            report_template
        )

        self._report_template = report_template
        self.data_source_descriptor = data_source_descriptor
        self._report_columns = report_columns

    @cached_property
    def filters_by_id(self):
        u"""Все фильтры шаблона по id фильтра.

        :rtype: dict
        """
        query = ReportFilter.objects.filter(
            group__report_template=self._report_template,
        )

        return {
            report_filter.id: report_filter
            for report_filter in query
        }

    def get_orm_filters(self):
        u"""Возвращает фильтры для формирования SQL-запроса.

        .. note::

           Предназначена для предварительной фильтрации данных на уровне СУБД.
           Т.к. данные загружаются только для модели источника данных без
           данных зависимых объектов, эти фильтры не срабатывают при загрузке
           зависимых объектов. Поэтому дополнительно нужно применять функцию
           фильтрации, возвращаемую методом :meth:`get_filter_function`.

        :rtype: django.db.models.Q
        """
        filter_group = self._report_template.filter_groups.filter(
            parent__isnull=True
        ).first()

        if filter_group:
            return _FilterGroupBuilder(self, filter_group).get_orm_filter()
        else:
            return Q()

    def _get_function_for_filter(self, report_filter):
        u"""Возвращает функцию для фильтрации данных в соответствии с фильтром.

        :rtype: callable
        """
        data_source = self.data_source_descriptor
        column_descriptor = data_source.get_column_descriptor(
            report_filter.column.name
        )
        filter_values = _get_filter_values(report_filter, column_descriptor)

        if report_filter.operator == constants.LE:
            def function(value):
                return value is not None and value <= filter_values

        elif report_filter.operator == constants.LT:
            def function(value):
                return value is not None and value < filter_values

        elif report_filter.operator == constants.EQ:
            def function(value):
                return value == filter_values

        elif report_filter.operator == constants.GT:
            def function(value):
                return value is not None and value > filter_values

        elif report_filter.operator == constants.GE:
            def function(value):
                return value is not None and value >= filter_values

        elif report_filter.operator == constants.IS_NULL:
            def function(value):
                return value is None or value == ''

        elif report_filter.operator == constants.CONTAINS:
            assert isinstance(filter_values, basestring), type(filter_values)

            if report_filter.case_sensitive:
                def function(value):
                    return filter_values in value
            else:
                def function(value):
                    return any(
                        v.lower() in value
                        for v in filter_values
                    )

        elif report_filter.operator == constants.STARTS_WITH:
            assert isinstance(filter_values, basestring), type(filter_values)

            if report_filter.case_sensitive:
                def function(value):
                    return value.startswith(filter_values)
            else:
                def function(value):
                    return value.lower().startswith(filter_values.lower())

        elif report_filter.operator == constants.ENDS_WITH:
            assert isinstance(filter_values, basestring), type(filter_values)

            if report_filter.case_sensitive:
                def function(value):
                    return value.endswith(filter_values)
            else:
                def function(value):
                    return value.lower().endswith(filter_values.lower())

        elif report_filter.operator == constants.BETWEEN:
            if report_filter.case_sensitive:
                def function(value):
                    return (
                        value is not None and
                        filter_values[0] <= value <= filter_values[1]
                    )
            else:
                def function(value):
                    if isinstance(value, unicode):
                        value = value.lower()
                        filter_values = tuple(v.lower() for v in filter_values)

                    return (
                        value is not None and
                        filter_values[0] <= value <= filter_values[1]
                    )

        elif report_filter.operator == constants.IN:
            assert isinstance(filter_values, tuple), type(filter_values)

            if report_filter.case_sensitive:
                def function(value):
                    return value in filter_values
            else:
                def function(value):
                    return value.lower() in (v.lower() for v in filter_values)

        else:
            raise FilterError(
                report_filter,
                u'Неподдерживаемый оператор ({})'.format(
                    report_filter.operator
                )
            )

        if report_filter.exclude:
            return lambda value: not function(value)
        else:
            return function

    @staticmethod
    def get_data_width(row_data):
        u"""Возвращает количество колонок, занимаемое данными.

        :param tuple row_data: Данные блока, строки или ячейки.
            - Ячейка --- это простые данные (строка, число, дата и т.п.).
            - Строка --- это кортеж из данных ячеек и вложенных блоков.
            - Блок --- это кортеж из строк. Содержит только строки.

        :rtype: int
        """
        is_row = DataLoader.is_row
        is_block = DataLoader.is_block
        get_data_width = _DataFilterer.get_data_width

        if is_block(row_data):
            return max(imap(get_data_width, row_data))
        elif is_row(row_data):
            return sum(imap(get_data_width, row_data))
        else:
            return 1

    def get_filter_function(self):
        u"""Возвращает функцию для фильтрации данных.

        Эта функция для каждого набора данных возвращает новый набор данных,
        которые отфильтрованы в соответствии с параметрами фильтрации,
        указанными в шаблоне отчета.

        Запись может быть отфильтрована полностью, либо отфильтрованы только
        вложенные блоки. В первом случае функция возвращает ``None``. Во
        втором случае --- будут удалены записи во внутренних блоках.
        """
        is_block = DataLoader.is_block
        get_data_width = _DataFilterer.get_data_width

        filter_functions = {
            report_filter.column_id: self._get_function_for_filter(
                report_filter
            )
            for report_filter in self.filters_by_id.itervalues()
        }
        filter_functions_by_column = tuple(
            filter_functions.get(report_column.pk)
            for report_column in self._report_columns
        )

        def filter_function(row_data, column_functions=None):
            u"""Возвращает отфильтрованные данные записи отчета."""
            if column_functions is None:
                column_functions = filter_functions_by_column

            i = 0
            result = []
            for cell in row_data:
                if is_block(cell):
                    cell_width = get_data_width(cell)
                    column_filter_functions = tuple(
                        column_functions[i + offset]
                        for offset in xrange(cell_width)
                    )
                    i += cell_width
                    filtered_cell_data = (
                        filter_function(cell_row_data, column_filter_functions)
                        for cell_row_data in cell
                    )
                    filtered_cell_data = tuple(
                        cell_row_data
                        for cell_row_data in filtered_cell_data
                        if cell_row_data is not None
                    )
                    if filtered_cell_data:
                        result.append(filtered_cell_data)
                    else:
                        return None
                else:
                    column_filter_function = column_functions[i]
                    if (
                        column_filter_function and
                        not column_filter_function(cell)
                    ):
                        return None
                    else:
                        result.append(cell)
                        i += 1

            return tuple(result)

        return filter_function


class _DataSorter(object):

    u"""Сортировщик данных отчета."""

    def __init__(self, report_template, data_source_descriptor):
        u"""Инициализация экземпляра класса.

        :param report_template: Шаблон фильтра.
        :type filter_group: educommon.report.constructor.models.ReportTemplate

        :param data_source_descriptor: Дескриптор источника данных..
        :type data_source_descriptor:
            educommon.report.constructor.base.ModelDataSourceDescriptor
        """
        assert isinstance(report_template, ReportTemplate), type(
            report_template
        )

        self._report_template = report_template
        self._data_source_descriptor = data_source_descriptor

    @cached_property
    def _report_sorting_params(self):
        u"""Параметры сортировки данных в шаблоне отчета.

        :rtype: tuple
        """
        return tuple(
            ReportSorting.objects.filter(
                column__report_template=self._report_template
            ).order_by('index').values_list(
                'column__name', 'direction'
            )
        )

    @cached_property
    def _column_descriptors(self):
        u"""Дескрипторы столбцов отчета, для которых определена сортировка.

        :rtype: dict
        """
        get_column_descriptor = (
            self._data_source_descriptor.get_column_descriptor
        )

        return {
            column_name: get_column_descriptor(column_name)
            for column_name, _ in self._report_sorting_params
        }

    def _get_lookup(self, column_name, direction):
        column_descriptor = self._column_descriptors[column_name]

        return '{}{}'.format(
            '-' if direction == constants.DIRECTION_DESC else '',
            column_descriptor.full_lookup,
        )

    def get_sorting_params(self):
        u"""Возвращает список полей для сортировки.

        :rtype: tuple
        """
        return tuple(
            self._get_lookup(column_name, direction)
            for column_name, direction in self._report_sorting_params
        )


class DataLoader(object):

    u"""Загрузчик данных для шаблона отчета.

    Данные каждой строки отчета представлены в виде кортежа. Элементами кортежа
    могут быть значения простых типов (числа, строки и т.п.), а также
    кортежи кортежей (вложенные блоки). Вложенные блоки добавляются в тех
    случаях, когда в параметрах отчета указаны т.н. обратные связи (связь "один
    ко многим"), когда одной записи в модели источника данных соответствует
    несколько записей в связанных моделях. В таких вложенных кортежах также
    могут быть простые данные и вложенные кортежи кортежей. Глубина вложенности
    не ограничивается и зависит от настроек шаблона отчета.

    Пример:

    .. table:: Данные в табличном виде
       :align: center

       +-------------+---------+---------------------------------------------+
       | **Фамилия** | **Имя** |                **Учащиеся**                 |
       |             |         +---------------------------------------------+
       |             |         |                 **Группа**                  |
       |             |         +--------------------------+------------------+
       |             |         |       **Учреждение**     | **Наименование** |
       |             |         +--------------------------+                  |
       |             |         | **Краткое наименование** |                  |
       +-------------+---------+--------------------------+------------------+
       | Иванов      | Иван    | СДЮШОР1                  | Бокс             |
       |             |         +--------------------------+------------------+
       |             |         | СДЮШОР1                  | Фехтование       |
       +-------------+---------+--------------------------+------------------+

    Одна запись в приведенной выше таблице будет иметь следующую структуру:

    .. code-block:: python

       (
           'Иванов',
           'Иван',
           (
               ('СДЮШОР1', 'Бокс'),
               ('СДЮШОР1', 'Фехтование'),
           )
       )
    """

    def __init__(self, report_template, data_source, report_columns, user):
        assert isinstance(report_template, ReportTemplate), type(
            report_template
        )

        self._report_template = report_template
        self._data_source = data_source
        self._report_columns = report_columns
        self._user = user

    @staticmethod
    def _extract(data_gen):
        result = tuple(data_gen)
        if not result:
            raise StopIteration()
        elif len(result) == 1:
            return result[0]
        else:
            return result

    @staticmethod
    def _get_column_count(columns):
        # Для иерархии столбцов возвращает количество занимаемых столбцов.
        if columns:
            return sum(imap(
                DataLoader._get_column_count, columns.itervalues()
            ))
        else:
            return 1

    @staticmethod
    def _get_field_display(field, field_value):
        for value, name in field.flatchoices:
            if value == field_value:
                return force_text(name, strings_only=True)

    @staticmethod
    def _get_object_data(obj, attr_name, nested):
        field = None
        if attr_name:
            field = get_field(obj, attr_name)
            model = field.model
            if isinstance(field, ForeignObjectRel):
                descriptor = getattr(model, field.get_accessor_name())
            elif isinstance(field, RelatedField):
                descriptor = getattr(model, field.name)
            else:
                descriptor = None

            if descriptor:
                try:
                    attr_value = getattr(obj, attr_name)
                except descriptor.RelatedObjectDoesNotExist:
                    attr_value = None
            else:
                attr_value = getattr(obj, attr_name)
        else:
            attr_value = obj
            descriptor = None

        # Обратная связь или M2M. Такие данные выводятся в виде кортежа.
        if (
            attr_name and
            (
                isinstance(field, ForeignObjectRel) and
                hasattr(descriptor, 'related_manager_cls') and
                isinstance(attr_value, descriptor.related_manager_cls) or
                isinstance(field, models.ManyToManyField) and
                isinstance(attr_value, Manager)
            )
        ):
            objects = attr_value.all()
            if objects:
                yield tuple(
                    DataLoader._extract(
                        DataLoader._get_object_data(related_obj, None, nested)
                    )
                    for related_obj in objects
                )
            else:
                yield ((None,) * DataLoader._get_column_count(nested),)

        # Прямая связь (ForeignKey или OneToOneField).
        elif nested:
            if attr_value:
                for k, v in nested.iteritems():
                    for object_data in (
                        DataLoader._get_object_data(attr_value, k, v)
                    ):
                        yield object_data
            else:
                for _ in xrange(DataLoader._get_column_count(nested)):
                    yield None

        # Поле с данными.
        else:
            if field.flatchoices:
                attr_value = DataLoader._get_field_display(field, attr_value)

            yield attr_value

    @staticmethod
    def is_row(data):
        u"""Возвращает True, если в ``data`` содержится строка с данными.

        :rtype: bool
        """
        return (
            isinstance(data, tuple) and
            any(not isinstance(cell, tuple) for cell in data)
        )

    @staticmethod
    def is_block(data):
        u"""Возвращает True, если в ``data`` содержится блок записей.

        :rtype: bool
        """
        return (
            isinstance(data, tuple) and
            all(isinstance(row, tuple) for row in data)
        )

    @cached_property
    def _data_filterer(self):
        u"""Фильтратор данных отчета."""
        return _DataFilterer(
            self._report_template, self._data_source, self._report_columns
        )

    def _available_units_filter(self, query):
        u"""Фильтратор данных отчета по доступности учреждений."""
        return self._data_source.add_source_filter(
            query, self._report_template.include_available_units,
            self._user
        )

    def _get_query_set(self):
        result = self._data_source.model.objects.all()
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Добавление фильтров в запрос.

        if ReportFilter.objects.filter(
            column__report_template=self._report_template,
        ).exists():
            result = result.filter(
                self._data_filterer.get_orm_filters()
            )
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Добавление параметров сортировки.

        result = result.order_by(
            *_DataSorter(
                self._report_template, self._data_source
            ).get_sorting_params()
        )
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Добавление параметров включения вложенных учреждений.

        result = self._available_units_filter(result)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        return result

    def _get_rows_data(self, columns_hierarchy):
        filter_function = self._data_filterer.get_filter_function()

        for obj in self._get_query_set():
            row_data = tuple(
                DataLoader._get_object_data(obj, None, columns_hierarchy)
            )
            if row_data and any(row_data):
                row_data = filter_function(row_data)
                if row_data:
                    yield row_data

    def __iter__(self):
        column_names = self._report_template.columns.filter(
            visible=True,
        ).values_list('name', flat=True)

        if not column_names:
            raise ReportConstructorException(
                u'В шаблоне нет ни одного отображаемого столбца.'
            )

        columns_hierarchy = get_columns_hierarchy(*column_names)

        return self._get_rows_data(columns_hierarchy)


class ReportBuilderBase(object):
    u"""Базовый класс построителя отчета."""

    __metaclass__ = abc.ABCMeta

    def __init__(self, report_template, file_path, user):
        u"""Инициализация класса.

        :param report_template: Шаблон отчета.
        :type report_template:
            educommon.report.constructor.models.ReportTemplate

        :param unicode file_path: Путь к файлу отчета.
        """
        assert isinstance(report_template, ReportTemplate), type(
            report_template
        )

        if report_template.data_source_name not in registry:
            raise DataSourceParamsNotFound(report_template.data_source_name)

        self._report_template = report_template
        self._file_path = file_path
        self._user = user

    @cached_property
    def _data_source(self):
        u"""Имя источника данных, указанного в шаблоне.

        :rtype: unicode
        """
        return self._report_template.data_source

    @cached_property
    def _report_columns(self):
        u"""Столбцы отчета.

        :rtype: tuple
        """
        return tuple(
            self._report_template.columns.all()
        )

    @cached_property
    def _data(self):
        u"""Отфильтрованные и отсортированные данные отчета.

        :rtype: tuple
        """
        return tuple(
            DataLoader(
                self._report_template, self._data_source, self._report_columns,
                self._user
            )
        )

    @cached_property
    def _workbook(self):
        u"""Книга Excel, в которой формируется отчет."""
        result = Workbook(self._file_path, dict(
            default_date_format='dd.mm.yyyy',
        ))

        for cell_format in (result.default_date_format,
                            result.default_url_format):
            cell_format.set_border()
            cell_format.set_align('vcenter')

        return result

    @cached_property
    def _worksheet(self):
        u"""Страница в книге Excel."""
        return self._workbook.add_worksheet(u'Отчет')

    @abc.abstractmethod
    def _flush_header(self):
        u"""Запись заголовка отчета."""

    @abc.abstractmethod
    def _flush_data(self):
        u"""Запись данных в отчет."""

    def build(self):
        u"""Формирование отчета."""
        self._flush_header()
        self._flush_data()

        self._workbook.close()
